/*
 * Project 64 - A Nintendo 64 emulator.
 *
 * (c) Copyright 2001 zilmar (zilmar@emulation64.com) and 
 * Jabo (jabo@emulation64.com).
 *
 * pj64 homepage: www.pj64.net
 *
 * Permission to use, copy, modify and distribute Project64 in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Project64 is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Project64 or software derived from Project64.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so if they want them.
 *
 */
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "main.h"
#include "cpu.h"
#include "debugger.h"
#include "plugin.h"
#include "al.h"

void * R4300i_Opcode[64];
void * R4300i_Special[64];
void * R4300i_Regimm[32];
void * R4300i_CoP0[32];
void * R4300i_CoP0_Function[64];
void * R4300i_CoP1[32];
void * R4300i_CoP1_BC[32];
void * R4300i_CoP1_S[64];
void * R4300i_CoP1_D[64];
void * R4300i_CoP1_W[64];
void * R4300i_CoP1_L[64];

FILE * loggy;

extern int shownewfunctioncalls;

char fcnused[0x1300000];

void _fastcall R4300i_opcode_SPECIAL (void) {
	((void (_fastcall *)()) R4300i_Special[ Opcode.funct ])();
}

void _fastcall R4300i_opcode_REGIMM (void) {
	((void (_fastcall *)()) R4300i_Regimm[ Opcode.rt ])();
}

void _fastcall R4300i_opcode_COP0 (void) {
	((void (_fastcall *)()) R4300i_CoP0[ Opcode.rs ])();
}

void _fastcall R4300i_opcode_COP0_CO (void) {
	((void (_fastcall *)()) R4300i_CoP0_Function[ Opcode.funct ])();
}

void _fastcall R4300i_opcode_COP1 (void) {
	((void (_fastcall *)()) R4300i_CoP1[ Opcode.fmt ])();
}

void _fastcall R4300i_opcode_COP1_BC (void) {
	((void (_fastcall *)()) R4300i_CoP1_BC[ Opcode.ft ])();
}

void _fastcall R4300i_opcode_COP1_S (void) {
	_controlfp(RoundingModel,_MCW_RC);
	((void (_fastcall *)()) R4300i_CoP1_S[ Opcode.funct ])();
}

void _fastcall R4300i_opcode_COP1_D (void) {
	_controlfp(RoundingModel,_MCW_RC);
	((void (_fastcall *)()) R4300i_CoP1_D[ Opcode.funct ])();
}

void _fastcall R4300i_opcode_COP1_W (void) {
	((void (_fastcall *)()) R4300i_CoP1_W[ Opcode.funct ])();
}

void _fastcall R4300i_opcode_COP1_L (void) {
	((void (_fastcall *)()) R4300i_CoP1_L[ Opcode.funct ])();
}

void BuildInterpreter (void ) { 
	R4300i_Opcode[ 0] = R4300i_opcode_SPECIAL;
	R4300i_Opcode[ 1] = R4300i_opcode_REGIMM;
	R4300i_Opcode[ 2] = r4300i_J;
	R4300i_Opcode[ 3] = r4300i_JAL;
	R4300i_Opcode[ 4] = r4300i_BEQ;
	R4300i_Opcode[ 5] = r4300i_BNE;
	R4300i_Opcode[ 6] = r4300i_BLEZ;
	R4300i_Opcode[ 7] = r4300i_BGTZ;
	R4300i_Opcode[ 8] = r4300i_ADDI;
	R4300i_Opcode[ 9] = r4300i_ADDIU;
	R4300i_Opcode[10] = r4300i_SLTI;
	R4300i_Opcode[11] = r4300i_SLTIU;
	R4300i_Opcode[12] = r4300i_ANDI;
	R4300i_Opcode[13] = r4300i_ORI;
	R4300i_Opcode[14] = r4300i_XORI;
	R4300i_Opcode[15] = r4300i_LUI;
	R4300i_Opcode[16] = R4300i_opcode_COP0;
	R4300i_Opcode[17] = R4300i_opcode_COP1;
	R4300i_Opcode[18] = R4300i_UnknownOpcode;
	R4300i_Opcode[19] = R4300i_UnknownOpcode;
	R4300i_Opcode[20] = r4300i_BEQL;
	R4300i_Opcode[21] = r4300i_BNEL;
	R4300i_Opcode[22] = r4300i_BLEZL;
	R4300i_Opcode[23] = r4300i_BGTZL;
	R4300i_Opcode[24] = R4300i_UnknownOpcode;
	R4300i_Opcode[25] = r4300i_DADDIU;
	R4300i_Opcode[26] = r4300i_LDL;
	R4300i_Opcode[27] = r4300i_LDR;
	R4300i_Opcode[28] = R4300i_UnknownOpcode;
	R4300i_Opcode[29] = R4300i_UnknownOpcode;
	R4300i_Opcode[30] = R4300i_UnknownOpcode;
	R4300i_Opcode[31] = R4300i_UnknownOpcode;
	R4300i_Opcode[32] = r4300i_LB;
	R4300i_Opcode[33] = r4300i_LH;
	R4300i_Opcode[34] = r4300i_LWL;
	R4300i_Opcode[35] = r4300i_LW;
	R4300i_Opcode[36] = r4300i_LBU;
	R4300i_Opcode[37] = r4300i_LHU;
	R4300i_Opcode[38] = r4300i_LWR;
	R4300i_Opcode[39] = r4300i_LWU;
	R4300i_Opcode[40] = r4300i_SB;
	R4300i_Opcode[41] = r4300i_SH;
	R4300i_Opcode[42] = r4300i_SWL;
	R4300i_Opcode[43] = r4300i_SW;
	R4300i_Opcode[44] = r4300i_SDL;
	R4300i_Opcode[45] = r4300i_SDR;
	R4300i_Opcode[46] = r4300i_SWR;
	R4300i_Opcode[47] = r4300i_CACHE;
	R4300i_Opcode[48] = r4300i_LL;
	R4300i_Opcode[49] = r4300i_LWC1;
	R4300i_Opcode[50] = R4300i_UnknownOpcode;
	R4300i_Opcode[51] = R4300i_UnknownOpcode;
	R4300i_Opcode[52] = R4300i_UnknownOpcode;
	R4300i_Opcode[53] = r4300i_LDC1;
	R4300i_Opcode[54] = R4300i_UnknownOpcode;
	R4300i_Opcode[55] = r4300i_LD;
	R4300i_Opcode[56] = r4300i_SC;
	R4300i_Opcode[57] = r4300i_SWC1;
	R4300i_Opcode[58] = R4300i_UnknownOpcode;
	R4300i_Opcode[59] = R4300i_UnknownOpcode;
	R4300i_Opcode[60] = R4300i_UnknownOpcode;
	R4300i_Opcode[61] = r4300i_SDC1;
	R4300i_Opcode[62] = R4300i_UnknownOpcode;
	R4300i_Opcode[63] = r4300i_SD;

	R4300i_Special[ 0] = r4300i_SPECIAL_SLL;
	R4300i_Special[ 1] = R4300i_UnknownOpcode;
	R4300i_Special[ 2] = r4300i_SPECIAL_SRL;
	R4300i_Special[ 3] = r4300i_SPECIAL_SRA;
	R4300i_Special[ 4] = r4300i_SPECIAL_SLLV;
	R4300i_Special[ 5] = R4300i_UnknownOpcode;
	R4300i_Special[ 6] = r4300i_SPECIAL_SRLV;
	R4300i_Special[ 7] = r4300i_SPECIAL_SRAV;
	R4300i_Special[ 8] = r4300i_SPECIAL_JR;
	R4300i_Special[ 9] = r4300i_SPECIAL_JALR;
	R4300i_Special[10] = R4300i_UnknownOpcode;
	R4300i_Special[11] = R4300i_UnknownOpcode;
	R4300i_Special[12] = r4300i_SPECIAL_SYSCALL;
	R4300i_Special[13] = R4300i_UnknownOpcode;
	R4300i_Special[14] = R4300i_UnknownOpcode;
	R4300i_Special[15] = r4300i_SPECIAL_SYNC;
	R4300i_Special[16] = r4300i_SPECIAL_MFHI;
	R4300i_Special[17] = r4300i_SPECIAL_MTHI;
	R4300i_Special[18] = r4300i_SPECIAL_MFLO;
	R4300i_Special[19] = r4300i_SPECIAL_MTLO;
	R4300i_Special[20] = r4300i_SPECIAL_DSLLV;
	R4300i_Special[21] = R4300i_UnknownOpcode;
	R4300i_Special[22] = r4300i_SPECIAL_DSRLV;
	R4300i_Special[23] = r4300i_SPECIAL_DSRAV;
	R4300i_Special[24] = r4300i_SPECIAL_MULT;
	R4300i_Special[25] = r4300i_SPECIAL_MULTU;
	R4300i_Special[26] = r4300i_SPECIAL_DIV;
	R4300i_Special[27] = r4300i_SPECIAL_DIVU;
	R4300i_Special[28] = r4300i_SPECIAL_DMULT;
	R4300i_Special[29] = r4300i_SPECIAL_DMULTU;
	R4300i_Special[30] = r4300i_SPECIAL_DDIV;
	R4300i_Special[31] = r4300i_SPECIAL_DDIVU;
	R4300i_Special[32] = r4300i_SPECIAL_ADD;
	R4300i_Special[33] = r4300i_SPECIAL_ADDU;
	R4300i_Special[34] = r4300i_SPECIAL_SUB;
	R4300i_Special[35] = r4300i_SPECIAL_SUBU;
	R4300i_Special[36] = r4300i_SPECIAL_AND;
	R4300i_Special[37] = r4300i_SPECIAL_OR;
	R4300i_Special[38] = r4300i_SPECIAL_XOR;
	R4300i_Special[39] = r4300i_SPECIAL_NOR;
	R4300i_Special[40] = R4300i_UnknownOpcode;
	R4300i_Special[41] = R4300i_UnknownOpcode;
	R4300i_Special[42] = r4300i_SPECIAL_SLT;
	R4300i_Special[43] = r4300i_SPECIAL_SLTU;
	R4300i_Special[44] = r4300i_SPECIAL_DADD;
	R4300i_Special[45] = r4300i_SPECIAL_DADDU;
	R4300i_Special[46] = r4300i_SPECIAL_DSUB;
	R4300i_Special[47] = r4300i_SPECIAL_DSUBU;
	R4300i_Special[48] = R4300i_UnknownOpcode;
	R4300i_Special[49] = R4300i_UnknownOpcode;
	R4300i_Special[50] = R4300i_UnknownOpcode;
	R4300i_Special[51] = R4300i_UnknownOpcode;
	R4300i_Special[52] = r4300i_SPECIAL_TEQ;
	R4300i_Special[53] = R4300i_UnknownOpcode;
	R4300i_Special[54] = R4300i_UnknownOpcode;
	R4300i_Special[55] = R4300i_UnknownOpcode;
	R4300i_Special[56] = r4300i_SPECIAL_DSLL;
	R4300i_Special[57] = R4300i_UnknownOpcode;
	R4300i_Special[58] = r4300i_SPECIAL_DSRL;
	R4300i_Special[59] = r4300i_SPECIAL_DSRA;
	R4300i_Special[60] = r4300i_SPECIAL_DSLL32;
	R4300i_Special[61] = R4300i_UnknownOpcode;
	R4300i_Special[62] = r4300i_SPECIAL_DSRL32;
	R4300i_Special[63] = r4300i_SPECIAL_DSRA32;

	R4300i_Regimm[ 0] = r4300i_REGIMM_BLTZ;
	R4300i_Regimm[ 1] = r4300i_REGIMM_BGEZ;
	R4300i_Regimm[ 2] = r4300i_REGIMM_BLTZL;
	R4300i_Regimm[ 3] = r4300i_REGIMM_BGEZL;
	R4300i_Regimm[ 4] = R4300i_UnknownOpcode;
	R4300i_Regimm[ 5] = R4300i_UnknownOpcode;
	R4300i_Regimm[ 6] = R4300i_UnknownOpcode;
	R4300i_Regimm[ 7] = R4300i_UnknownOpcode;
	R4300i_Regimm[ 8] = R4300i_UnknownOpcode;
	R4300i_Regimm[ 9] = R4300i_UnknownOpcode;
	R4300i_Regimm[10] = R4300i_UnknownOpcode;
	R4300i_Regimm[11] = R4300i_UnknownOpcode;
	R4300i_Regimm[12] = R4300i_UnknownOpcode;
	R4300i_Regimm[13] = R4300i_UnknownOpcode;
	R4300i_Regimm[14] = R4300i_UnknownOpcode;
	R4300i_Regimm[15] = R4300i_UnknownOpcode;
	R4300i_Regimm[16] = r4300i_REGIMM_BLTZAL;
	R4300i_Regimm[17] = r4300i_REGIMM_BGEZAL;
	R4300i_Regimm[18] = R4300i_UnknownOpcode;
	R4300i_Regimm[19] = R4300i_UnknownOpcode;
	R4300i_Regimm[20] = R4300i_UnknownOpcode;
	R4300i_Regimm[21] = R4300i_UnknownOpcode;
	R4300i_Regimm[22] = R4300i_UnknownOpcode;
	R4300i_Regimm[23] = R4300i_UnknownOpcode;
	R4300i_Regimm[24] = R4300i_UnknownOpcode;
	R4300i_Regimm[25] = R4300i_UnknownOpcode;
	R4300i_Regimm[26] = R4300i_UnknownOpcode;
	R4300i_Regimm[27] = R4300i_UnknownOpcode;
	R4300i_Regimm[28] = R4300i_UnknownOpcode;
	R4300i_Regimm[29] = R4300i_UnknownOpcode;
	R4300i_Regimm[30] = R4300i_UnknownOpcode;
	R4300i_Regimm[31] = R4300i_UnknownOpcode;
	
	R4300i_CoP0[ 0] = r4300i_COP0_MF;
	R4300i_CoP0[ 1] = R4300i_UnknownOpcode;
	R4300i_CoP0[ 2] = R4300i_UnknownOpcode;
	R4300i_CoP0[ 3] = R4300i_UnknownOpcode;
	R4300i_CoP0[ 4] = r4300i_COP0_MT;
	R4300i_CoP0[ 5] = R4300i_UnknownOpcode;
	R4300i_CoP0[ 6] = R4300i_UnknownOpcode;
	R4300i_CoP0[ 7] = R4300i_UnknownOpcode;
	R4300i_CoP0[ 8] = R4300i_UnknownOpcode;
	R4300i_CoP0[ 9] = R4300i_UnknownOpcode;
	R4300i_CoP0[10] = R4300i_UnknownOpcode;
	R4300i_CoP0[11] = R4300i_UnknownOpcode;
	R4300i_CoP0[12] = R4300i_UnknownOpcode;
	R4300i_CoP0[13] = R4300i_UnknownOpcode;
	R4300i_CoP0[14] = R4300i_UnknownOpcode;
	R4300i_CoP0[15] = R4300i_UnknownOpcode;
	R4300i_CoP0[16] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[17] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[18] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[19] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[20] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[21] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[22] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[23] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[24] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[25] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[26] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[27] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[28] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[29] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[30] = R4300i_opcode_COP0_CO;
	R4300i_CoP0[31] = R4300i_opcode_COP0_CO;

	R4300i_CoP0_Function[ 0] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[ 1] = r4300i_COP0_CO_TLBR;
	R4300i_CoP0_Function[ 2] = r4300i_COP0_CO_TLBWI;
	R4300i_CoP0_Function[ 3] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[ 4] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[ 5] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[ 6] = r4300i_COP0_CO_TLBWR;
	R4300i_CoP0_Function[ 7] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[ 8] = r4300i_COP0_CO_TLBP;
	R4300i_CoP0_Function[ 9] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[10] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[11] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[12] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[13] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[14] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[15] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[16] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[17] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[18] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[19] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[20] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[21] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[22] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[23] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[24] = r4300i_COP0_CO_ERET;
	R4300i_CoP0_Function[25] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[26] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[27] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[28] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[29] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[30] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[31] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[32] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[33] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[34] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[35] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[36] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[37] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[38] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[39] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[40] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[41] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[42] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[43] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[44] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[45] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[46] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[47] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[48] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[49] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[50] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[51] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[52] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[53] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[54] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[55] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[56] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[57] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[58] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[59] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[60] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[61] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[62] = R4300i_UnknownOpcode;
	R4300i_CoP0_Function[63] = R4300i_UnknownOpcode;
	
	R4300i_CoP1[ 0] = r4300i_COP1_MF;
	R4300i_CoP1[ 1] = r4300i_COP1_DMF;
	R4300i_CoP1[ 2] = r4300i_COP1_CF;
	R4300i_CoP1[ 3] = R4300i_UnknownOpcode;
	R4300i_CoP1[ 4] = r4300i_COP1_MT;
	R4300i_CoP1[ 5] = r4300i_COP1_DMT;
	R4300i_CoP1[ 6] = r4300i_COP1_CT;
	R4300i_CoP1[ 7] = R4300i_UnknownOpcode;
	R4300i_CoP1[ 8] = R4300i_opcode_COP1_BC;
	R4300i_CoP1[ 9] = R4300i_UnknownOpcode;
	R4300i_CoP1[10] = R4300i_UnknownOpcode;
	R4300i_CoP1[11] = R4300i_UnknownOpcode;
	R4300i_CoP1[12] = R4300i_UnknownOpcode;
	R4300i_CoP1[13] = R4300i_UnknownOpcode;
	R4300i_CoP1[14] = R4300i_UnknownOpcode;
	R4300i_CoP1[15] = R4300i_UnknownOpcode;
	R4300i_CoP1[16] = R4300i_opcode_COP1_S;
	R4300i_CoP1[17] = R4300i_opcode_COP1_D;
	R4300i_CoP1[18] = R4300i_UnknownOpcode;
	R4300i_CoP1[19] = R4300i_UnknownOpcode;
	R4300i_CoP1[20] = R4300i_opcode_COP1_W;
	R4300i_CoP1[21] = R4300i_opcode_COP1_L;
	R4300i_CoP1[22] = R4300i_UnknownOpcode;
	R4300i_CoP1[23] = R4300i_UnknownOpcode;
	R4300i_CoP1[24] = R4300i_UnknownOpcode;
	R4300i_CoP1[25] = R4300i_UnknownOpcode;
	R4300i_CoP1[26] = R4300i_UnknownOpcode;
	R4300i_CoP1[27] = R4300i_UnknownOpcode;
	R4300i_CoP1[28] = R4300i_UnknownOpcode;
	R4300i_CoP1[29] = R4300i_UnknownOpcode;
	R4300i_CoP1[30] = R4300i_UnknownOpcode;
	R4300i_CoP1[31] = R4300i_UnknownOpcode;

	R4300i_CoP1_BC[ 0] = r4300i_COP1_BCF;
	R4300i_CoP1_BC[ 1] = r4300i_COP1_BCT;
	R4300i_CoP1_BC[ 2] = r4300i_COP1_BCFL;
	R4300i_CoP1_BC[ 3] = r4300i_COP1_BCTL;
	R4300i_CoP1_BC[ 4] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[ 5] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[ 6] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[ 7] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[ 8] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[ 9] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[10] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[11] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[12] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[13] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[14] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[15] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[16] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[17] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[18] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[19] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[20] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[21] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[22] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[23] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[24] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[25] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[26] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[27] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[28] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[29] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[30] = R4300i_UnknownOpcode;
	R4300i_CoP1_BC[31] = R4300i_UnknownOpcode;

	R4300i_CoP1_S[ 0] = r4300i_COP1_S_ADD;
	R4300i_CoP1_S[ 1] = r4300i_COP1_S_SUB;
	R4300i_CoP1_S[ 2] = r4300i_COP1_S_MUL;
	R4300i_CoP1_S[ 3] = r4300i_COP1_S_DIV;
	R4300i_CoP1_S[ 4] = r4300i_COP1_S_SQRT;
	R4300i_CoP1_S[ 5] = r4300i_COP1_S_ABS;
	R4300i_CoP1_S[ 6] = r4300i_COP1_S_MOV;
	R4300i_CoP1_S[ 7] = r4300i_COP1_S_NEG;
	R4300i_CoP1_S[ 8] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[ 9] = r4300i_COP1_S_TRUNC_L;
	R4300i_CoP1_S[10] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[11] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[12] = r4300i_COP1_S_ROUND_W;
	R4300i_CoP1_S[13] = r4300i_COP1_S_TRUNC_W;
	R4300i_CoP1_S[14] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[15] = r4300i_COP1_S_FLOOR_W;
	R4300i_CoP1_S[16] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[17] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[18] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[19] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[20] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[21] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[22] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[23] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[24] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[25] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[26] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[27] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[28] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[29] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[30] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[31] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[32] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[33] = r4300i_COP1_S_CVT_D;
	R4300i_CoP1_S[34] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[35] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[36] = r4300i_COP1_S_CVT_W;
	R4300i_CoP1_S[37] = r4300i_COP1_S_CVT_L;
	R4300i_CoP1_S[38] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[39] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[40] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[41] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[42] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[43] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[44] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[45] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[46] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[47] = R4300i_UnknownOpcode;
	R4300i_CoP1_S[48] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[49] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[50] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[51] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[52] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[53] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[54] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[55] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[56] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[57] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[58] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[59] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[60] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[61] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[62] = r4300i_COP1_S_CMP;
	R4300i_CoP1_S[63] = r4300i_COP1_S_CMP;

	R4300i_CoP1_D[ 0] = r4300i_COP1_D_ADD;
	R4300i_CoP1_D[ 1] = r4300i_COP1_D_SUB;
	R4300i_CoP1_D[ 2] = r4300i_COP1_D_MUL;
	R4300i_CoP1_D[ 3] = r4300i_COP1_D_DIV;
	R4300i_CoP1_D[ 4] = r4300i_COP1_D_SQRT;
	R4300i_CoP1_D[ 5] = r4300i_COP1_D_ABS;
	R4300i_CoP1_D[ 6] = r4300i_COP1_D_MOV;
	R4300i_CoP1_D[ 7] = r4300i_COP1_D_NEG;
	R4300i_CoP1_D[ 8] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[ 9] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[10] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[11] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[12] = r4300i_COP1_D_ROUND_W;
	R4300i_CoP1_D[13] = r4300i_COP1_D_TRUNC_W;
	R4300i_CoP1_D[14] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[15] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[16] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[17] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[18] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[19] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[20] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[21] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[22] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[23] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[24] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[25] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[26] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[27] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[28] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[29] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[30] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[31] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[32] = r4300i_COP1_D_CVT_S;
	R4300i_CoP1_D[33] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[34] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[35] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[36] = r4300i_COP1_D_CVT_W;
	R4300i_CoP1_D[37] = r4300i_COP1_D_CVT_L;
	R4300i_CoP1_D[38] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[39] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[40] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[41] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[42] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[43] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[44] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[45] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[46] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[47] = R4300i_UnknownOpcode;
	R4300i_CoP1_D[48] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[49] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[50] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[51] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[52] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[53] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[54] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[55] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[56] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[57] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[58] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[59] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[60] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[61] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[62] = r4300i_COP1_D_CMP;
	R4300i_CoP1_D[63] = r4300i_COP1_D_CMP;

	R4300i_CoP1_W[ 0] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[ 1] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[ 2] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[ 3] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[ 4] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[ 5] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[ 6] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[ 7] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[ 8] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[ 9] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[10] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[11] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[12] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[13] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[14] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[15] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[16] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[17] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[18] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[19] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[20] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[21] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[22] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[23] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[24] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[25] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[26] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[27] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[28] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[29] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[30] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[31] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[32] = r4300i_COP1_W_CVT_S;
	R4300i_CoP1_W[33] = r4300i_COP1_W_CVT_D;
	R4300i_CoP1_W[34] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[35] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[36] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[37] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[38] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[39] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[40] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[41] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[42] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[43] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[44] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[45] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[46] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[47] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[48] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[49] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[50] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[51] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[52] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[53] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[54] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[55] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[56] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[57] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[58] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[59] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[60] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[61] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[62] = R4300i_UnknownOpcode;
	R4300i_CoP1_W[63] = R4300i_UnknownOpcode;	

	R4300i_CoP1_L[ 0] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[ 1] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[ 2] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[ 3] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[ 4] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[ 5] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[ 6] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[ 7] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[ 8] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[ 9] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[10] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[11] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[12] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[13] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[14] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[15] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[16] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[17] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[18] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[19] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[20] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[21] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[22] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[23] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[24] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[25] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[26] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[27] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[28] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[29] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[30] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[31] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[32] = r4300i_COP1_L_CVT_S;
	R4300i_CoP1_L[33] = r4300i_COP1_L_CVT_D;
	R4300i_CoP1_L[34] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[35] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[36] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[37] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[38] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[39] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[40] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[41] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[42] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[43] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[44] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[45] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[46] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[47] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[48] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[49] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[50] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[51] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[52] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[53] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[54] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[55] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[56] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[57] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[58] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[59] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[60] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[61] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[62] = R4300i_UnknownOpcode;
	R4300i_CoP1_L[63] = R4300i_UnknownOpcode;	
}

DWORD TLBTrans(DWORD in) {
	return (TLB_ReadMap[in>>12]+(BYTE*)in-N64MEM);
}

void ShowCPUStatus(void) {
	DisplayError("r0=%08x at=%08x v0=%08x v1=%08x a0=%08x a1=%08x a2=%08x a3=%08x\n"
				 "t0=%08x t1=%08x t2=%08x t3=%08x t4=%08x t5=%08x t6=%08x t7=%08x\n"
				 "s0=%08x s1=%08x s2=%08x s3=%08x s4=%08x s5=%08x s6=%08x s7=%08x\n"
                 "t8=%08x t9=%08x k0=%08x k1=%08x gp=%08x sp=%08x s8=%08x ra=%08x (%08x)\nPC=%08x",
				  GPR[0].UW[0],GPR[1].UW[0],GPR[2].UW[0],GPR[3].UW[0],GPR[4].UW[0],GPR[5].UW[0],GPR[6].UW[0],GPR[7].UW[0],
				  GPR[8].UW[0],GPR[9].UW[0],GPR[10].UW[0],GPR[11].UW[0],GPR[12].UW[0],GPR[13].UW[0],GPR[14].UW[0],GPR[15].UW[0],
				  GPR[16].UW[0],GPR[17].UW[0],GPR[18].UW[0],GPR[19].UW[0],GPR[20].UW[0],GPR[21].UW[0],GPR[22].UW[0],GPR[23].UW[0],
				  GPR[24].UW[0],GPR[25].UW[0],GPR[26].UW[0],GPR[27].UW[0],GPR[28].UW[0],GPR[29].UW[0],GPR[30].UW[0],GPR[31].UW[0],
				  TLBTrans(GPR[31].UW[0]),PROGRAM_COUNTER);
}

void ExecuteInterpreterOpCode (void) {
	static int used=0;
	static FILE * dmalog=NULL;
	int c=0;
	DWORD cp;
	//static char doneyet[0x400000/4]={0};

	// Cleanup for Mystical Ninja
	/*if (PROGRAM_COUNTER==0x80038BC8) {
		GPR[4].UW[0]=0x20;
		((DWORD*)N64MEM)[0xcc0/4]=0;
		sprintf(SaveAsFileName,"HITME");
		Machine_SaveState();
	}*/

	// Cleanup for Wave Race
	/*if (PROGRAM_COUNTER==0x800c56d4) {
		((DWORD*)N64MEM)[0x471e0/4]=0x08011c70;
	
		sprintf(SaveAsFileName,"WAVERACE_HITME");
		Machine_SaveState();
	}*/

	// Cleanup for Blast Corps
	/*if (PROGRAM_COUNTER==0x80244930) {
		((DWORD*)N64MEM)[0x24494c/4]=0x0c0987EE; // jal 0x80261FB8
		((DWORD*)N64MEM)[0x244950/4]=0x20040011; // li a0,0x11
		((DWORD*)N64MEM)[0x244954/4]=0x08091D6E; // j 0x802475b8
		((DWORD*)N64MEM)[0x244958/4]=0;

		sprintf(SaveAsFileName,"BCORPS_HITME");
		Machine_SaveState();
	}*/

	
	// GGA
	/*if (PROGRAM_COUNTER==0x80000d00) DisplayError("hiya");
	if (PROGRAM_COUNTER==0x80008068) {
		//DisplayError("a0=%08x a1=%08x a2=%08x\nra=%08x",GPR[4].UW[0],GPR[5].UW[0],GPR[6].UW[0],GPR[31].UW[0]);
		GPR[4].UW[0]=0x56;
		((DWORD*)N64MEM)[0x1224/4]=0;
		((DWORD*)N64MEM)[0x1214/4]=0;
		((DWORD*)N64MEM)[0x11EC/4]=0;
		((DWORD*)N64MEM)[0x11cc/4]=0;  // all four needed to kill Thread6
		((DWORD*)N64MEM)[0xcf4/4]=0;  // kill Thread4 (video)
		((DWORD*)N64MEM)[0x1364/4]=0; // kill Thread7 (game logic)
		((DWORD*)N64MEM)[0xD1F08/4]=0; // 
		((DWORD*)N64MEM)[0x1E768/4]=0; // Thread10
		//((DWORD*)N64MEM)[0x16CCC/4]=0; //Thread9
		
		//((DWORD*)N64MEM)[0x7590/4]=0x34060000; // make audio thread blocking
		//used=1;
		//((DWORD*)N64MEM)[0x16F190]=0x08000000|(0x16F30C>>2);
		//((DWORD*)N64MEM)[0xC5984/4]=0x0803171B;

		sprintf(SaveAsFileName,"GGA");
		Machine_SaveState();
	}
	//if (PROGRAM_COUNTER==0x8000758C && used) GPR[6].UW[0]=1;
	*/


	// DK64
	//if (PROGRAM_COUNTER==0x80024000 && used) PROGRAM_COUNTER=GPR[31].UW[0];
	//if (PROGRAM_COUNTER==0x805FC2B0 && used) PROGRAM_COUNTER=GPR[31].UW[0];
	
	/*if (PROGRAM_COUNTER==0x800049F0) DisplayError("Thread at %08x",GPR[6].UW[0]);
	if (PROGRAM_COUNTER==0x8060E174) {
		//DisplayError("HO!\n%08x",GPR[25].UW[0]);
		((DWORD*)N64MEM)[0x5FC270/4]=0;
		//((DWORD*)N64MEM)[0x5FC278/4]=0x0c000000|(0x602A94>>2);
		//((WORD*)N64MEM)[0x7458DC/2]=0x3;
		//used=1;
	}

	if (PROGRAM_COUNTER==0x805FC278) {
		sprintf(SaveAsFileName,"DK64_HITME");
		Machine_SaveState();
	}*/

	/*if (PROGRAM_COUNTER==0x805fed68) {
		PROGRAM_COUNTER=0x805FEE70;
		//((DWORD*)N64MEM)[0x5FC270/4]=0;
		((DWORD*)N64MEM)[0x5FC0D0/4]=0; // disable calling stage fcn again
		//((DWORD*)N64MEM)[0x60F184/4]=0; // disable SP?
		((DWORD*)N64MEM)[0x732330/4]=0; // disable Thread7 (useless)

		//((DWORD*)N64MEM)[0x4F58/4]=0; // disable Thread2 (kills audio)

		//((DWORD*)N64MEM)[0xC46C/4]=0; // diable Thread4 (kills dlist, but no progress)

		//((DWORD*)N64MEM)[0x60E17C/4]=0; // disable Thread8
		
		((DWORD*)N64MEM)[0x24000/4]=0x03E00008;
		((DWORD*)N64MEM)[0x24004/4]=0;

		//((DWORD*)N64MEM)[0x5FC2B0/4]=0x03E00008;
		//((DWORD*)N64MEM)[0x5FC2B4/4]=0;
		//((DWORD*)N64MEM)[0x5FC164/4]=0;
		//((DWORD*)N64MEM)[0x5FC16C/4]=0;
		//((DWORD*)N64MEM)[0x5FC17C/4]=0;
		//((DWORD*)N64MEM)[0x5FC194/4]=0;
		//((DWORD*)N64MEM)[0x5FC19C/4]=0;
		//((DWORD*)N64MEM)[0x5FC1C4/4]=0;
		//((DWORD*)N64MEM)[0x5FC1CC/4]=0;
		//((DWORD*)N64MEM)[0x5FC1D4/4]=0;
		//((DWORD*)N64MEM)[0x5FC208/4]=0;
		//((DWORD*)N64MEM)[0x5FC224/4]=0;
	}*/

	
	/*
	// This bit *almost* works
	if (PROGRAM_COUNTER==0x80602A94) {
		GPR[4].UW[0]=0xae;
		//DisplayError("a0=%08x a1=%08x ra=%08x",GPR[4].UW[0],GPR[5].UW[0],GPR[31].UW[0]);

		((DWORD*)N64MEM)[0x732330/4]=0; // disable Thread7 (useless)

		//((DWORD*)N64MEM)[0xC46C/4]=0; // diable Thread4 (kills dlist, but no progress)

		((DWORD*)N64MEM)[0x60E17C/4]=0; // disable Thread8

		//((DWORD*)N64MEM)[0x60F18C/4]=0; // disable Thread5

		((DWORD*)N64MEM)[0x5FC270/4]=0; // disable ThreadMain

		//((DWORD*)N64MEM)[0x4F48/4]=0; // disable Thread3
		//((DWORD*)N64MEM)[0x4F50/4]=0;		

		((DWORD*)N64MEM)[0x5fed68/4]=0x08000000|(0x5FEE70>>2);
		((DWORD*)N64MEM)[0x5fed6c/4]=0;

		((DWORD*)N64MEM)[0x5FC0D0/4]=0; // disable calling stage fcn again

		((DWORD*)N64MEM)[0x24000/4]=0x03E00008;
		((DWORD*)N64MEM)[0x24004/4]=0;

		// disable SI stuff
		//((DWORD*)N64MEM)[0xAE20/4]=0x03E00008;
		//((DWORD*)N64MEM)[0xAE24/4]=0;
		//((DWORD*)N64MEM)[0x7120/4]=0;
		//((DWORD*)N64MEM)[0x7144/4]=0;
		//((DWORD*)N64MEM)[0x7414/4]=0;
		//((DWORD*)N64MEM)[0x7a38/4]=0;
		//((DWORD*)N64MEM)[0x7a64/4]=0;
		//((DWORD*)N64MEM)[0x7c44/4]=0;
		//((DWORD*)N64MEM)[0x7c74/4]=0;
		//((DWORD*)N64MEM)[0x7e5c/4]=0;
		//((DWORD*)N64MEM)[0x7e88/4]=0;
		//((DWORD*)N64MEM)[0xb07c/4]=0;
		//((DWORD*)N64MEM)[0xb0a8/4]=0;
		//((DWORD*)N64MEM)[0xb51c/4]=0;
		//((DWORD*)N64MEM)[0xb538/4]=0;
		//((DWORD*)N64MEM)[0xc60c/4]=0;
		//((DWORD*)N64MEM)[0xc634/4]=0;
		//((DWORD*)N64MEM)[0xc728/4]=0;
		//((DWORD*)N64MEM)[0xc74c/4]=0;
		//((DWORD*)N64MEM)[0xca60/4]=0;
		//((DWORD*)N64MEM)[0xca84/4]=0;

		((DWORD*)N64MEM)[0xA3A0/4]=0x03E00008;
		((DWORD*)N64MEM)[0xA3A4/4]=0;

		((DWORD*)N64MEM)[0xA3F0/4]=0x03E00008;
		((DWORD*)N64MEM)[0xA3F4/4]=0;

		((DWORD*)N64MEM)[0x60F168/4]=0;
		//((DWORD*)N64MEM)[0x60F490/4]=0; // kills alist
		//((DWORD*)N64MEM)[0x60F838/4]=0; // kills everything

		//((DWORD*)N64MEM)[0x60F418/4]=0;

		// try killing some osSendMesgs
		//((DWORD*)N64MEM)[0x4EF8/4]=0; // kills alist
		
		((DWORD*)N64MEM)[0xC254/4]=0;
		((DWORD*)N64MEM)[0xC294/4]=0;
		((DWORD*)N64MEM)[0xC408/4]=0;
		
		//((DWORD*)N64MEM)[0xC450/4]=0; // <<<--- responsible for freezing everything
		//((DWORD*)N64MEM)[0xC464/4]=0; // <<<--- ditto

		((DWORD*)N64MEM)[0x5FC224/4]=0;
		((DWORD*)N64MEM)[0x601D98/4]=0;

		//((DWORD*)N64MEM)[0x601E04/4]=0; // no alists

		//((DWORD*)N64MEM)[0x6020D0/4]=0;

		((DWORD*)N64MEM)[0x60B950/4]=0;

		((DWORD*)N64MEM)[0x60F168/4]=0;

		//((DWORD*)N64MEM)[0x60F490/4]=0;

		((DWORD*)N64MEM)[0x60F51C/4]=0;

		//((DWORD*)N64MEM)[0x60F8FC/4]=0;

		//((DWORD*)N64MEM)[0x60FACC/4]=0; // hmmm

		((BYTE*)N64MEM)[0x745650^3]=1; // aha! "do sound NOW"
		
		((DWORD*)N64MEM)[0x60FACC/4]=0; 

		//used=1;

		//DlistCount=0;

		sprintf(SaveAsFileName,"DK64_MIGHTY");
		Machine_SaveState();
	}*/

	//if (used && PROGRAM_COUNTER==0x8060FACC) {
	//if (used && PROGRAM_COUNTER==0x80602630) {
	//	((DWORD*)N64MEM)[0x60FACC/4]=0; 
	//	//else used++;
	//	
	//}
	//if (PROGRAM_COUNTER==0x80602B00) DisplayError("a3=%08x",GPR[7].UW[0]);

	// Goemon - Mononoke Sugoroku
	/*
	if (PROGRAM_COUNTER==0x80017D4C) {
		((DWORD*)N64MEM)[0xf28/4]=0; // kill Thread6
		((DWORD*)N64MEM)[0xf14/4]=0;
		((DWORD*)N64MEM)[0xf04/4]=0;
		((DWORD*)N64MEM)[0xecc/4]=0;
		((DWORD*)N64MEM)[0xebc/4]=0;

		((DWORD*)N64MEM)[0xf24/4]=0x8fbf002c;
		((DWORD*)N64MEM)[0xf28/4]=0x03e00008;
		((DWORD*)N64MEM)[0xf2c/4]=0x27bd0030;
		
		//((DWORD*)N64MEM)[0xea0/4]=0x8fb20028;
		//((DWORD*)N64MEM)[0xea4/4]=0x8fb10024;
		//((DWORD*)N64MEM)[0xea8/4]=0x8fb00020;
		
		//((DWORD*)N64MEM)[0xe9c/4]=0x0c000000|(0xf30>>2);
		//((DWORD*)N64MEM)[0xea0/4]=0;
		
		//((DWORD*)N64MEM)[0x22aa0/4]=0; // kill Thread8

		//((WORD*)N64MEM)[0x173A8/2]=0; // make audio blocking

		//((DWORD*)N64MEM)[0xBA4/4]=0; // kill Thread4

		if (GPR[4].UW[0]==0x0a) {
			DisplayError("a0=%08x ra=%08x",GPR[4].UW[0],GPR[31].UW[0]);
			GPR[4].UW[0]=0x39;
		
			sprintf(SaveAsFileName,"MONO");
			Machine_SaveState();
		}
	}
	*/

	/*if (PROGRAM_COUNTER==0x80000400) {
		sprintf(SaveAsFileName,"JFGOMG");
		Machine_SaveState();
	}*/

	//if (PROGRAM_COUNTER==0x80007EC8) DisplayError("a0=%08x ra=%08x",GPR[4].UW[0],GPR[31].UW[0]);

	//if (PROGRAM_COUNTER==0x800c7ea0) DisplayError("a0=%08x a1=%08x a2=%08x",GPR[4].UW[0],GPR[5].UW[0],GPR[6].UW[0]);

	#ifdef AL_USE

	if (PROGRAM_COUNTER != 0 && PROGRAM_COUNTER==alGetAddress("alBnkfNew")) {
		DWORD ctl_offset = GPR[4].UW[0];
		DWORD tbl_offset = GPR[5].UW[0];

		// try to determine ctl size
		unsigned long ctl_size = compute_ctl_size(ctl_offset);

		char ctl_report[512] = "";
		BYTE * ctl = malloc(ctl_size);
		
		unsigned long offset = 0;
		unsigned long i;
		for (i = 0; i < ctl_size; i++) {
			r4300i_LB_VAddr(ctl_offset+i, &ctl[i]);
		}
		
		for (offset = 0; offset < RomFileSize; offset += 4)
		{
			for (i = 0; i < ctl_size; i++)
			{
				if (ROM[(offset+i)^3] != ctl[i]) break;
			}
			
			if (i == ctl_size)
			{
				int len = strlen(ctl_report);
				sprintf_s(ctl_report+len, sizeof(ctl_report)-len-1, "%s%08x", (len > 0 ? ", " : ""), offset);
			}
		}

		free(ctl);

		DisplayError("alBnkfNew\nctl=%08x (size %08lx: %s)\ntbl=%08x",ctl_offset, ctl_size, ctl_report, tbl_offset);
	}

#if 0
	/*if (PROGRAM_COUNTER==alGetAddress("alSeqFileNew")) {
		DisplayError("alSeqFileNew\naddress=%08x",GPR[5].UW[0]);
		
	}

	if (PROGRAM_COUNTER==alGetAddress("alCSeqNew")) {
		DisplayError("alCSeqNew\naddress=%08x",GPR[5].UW[0]);
		
		//dumpseq(GPR[5].UW[0]);
	}*/
	
	//if (PROGRAM_COUNTER==alGetAddress("osAiSetFrequency")) {
	//	DisplayError("osAiSetFrequency\nfrequency=%d",GPR[4].UW[0]);
	//}
	/*if (PROGRAM_COUNTER==alGetAddress("alInit")) {
		// a1 = synconfig
		DWORD * synconfig=((DWORD*)N64MEM)+(GPR[5].UW[0]&0x7fffff)/4;
		DisplayError("alInit\nmaxVVoices=%08x\nmaxPVoices=%08x\nmaxUpdates=%08x\nmaxFXbusses=%08x\nfxType=%02x\nparams=%08x",
			synconfig[0],synconfig[1],synconfig[2],synconfig[3],((BYTE*)synconfig)[(7*4+0)^3],synconfig[8]);

		if (((BYTE*)synconfig)[(7*4+0)^3]==6) { // custom fx
			FILE * params = fopen("h:\\fxcustom.txt","w");
			DWORD c,t,a=synconfig[8];
			r4300i_LW_VAddr(a,&c);
			r4300i_LW_VAddr(a+4,&t);
			fprintf(params,"{%ld, ",c); // count
			fprintf(params,"0x%lx,\n",t);

			a+=8;
			c*=8;
			c--;
			for (; c >= 0; c--,a+=4) {
				r4300i_LW_VAddr(a,&t);
				fprintf(params,"0x%x, ",t);
				if ((c%8)==0) fprintf(params,"\n");
				if (!c) break;
			}
			fprintf(params,"}");
			fclose(params);
		}
	}
	if (PROGRAM_COUNTER==alGetAddress("n_alInit")) {
		DisplayError("n_alInit");
	}

	if (PROGRAM_COUNTER==alGetAddress("alBnkfNew")) {
		DisplayError("alBnkfNew\nctl=%08x\ntbl=%08x",GPR[4].UW[0],GPR[5].UW[0]);
	}

	if (PROGRAM_COUNTER==alGetAddress("alCSPNew")) {
		DWORD * seqconfig=((DWORD*)N64MEM)+(GPR[5].UW[0]&0x7fffff)/4;
		DisplayError("alCSPNew\nmaxVoices=%08x\nmaxEvents=%08x\nmaxChannels=%02x\ndebugFlags=%02x",seqconfig[0],seqconfig[1],((BYTE*)seqconfig)[(2*4+0)^3],((BYTE*)seqconfig)[(2*4+1)^3]);
		if (seqconfig[4] || seqconfig[5] || seqconfig[6]) DisplayError("custom osc fcns");
	}*/
#endif
#endif


	//if (PROGRAM_COUNTER==0x8004D530 && GPR[4].UW[0]!=0x1f && GPR[4].UW[0] != 0x3b) DisplayError("a0=%08x ra=%08x",GPR[4].UW[0],GPR[31].UW[0]);

	// JFG's alInit
	/*if (PROGRAM_COUNTER==0x8008A540) {
		// a1 = synconfig
		DWORD * synconfig=((DWORD*)N64MEM)+(GPR[5].UW[0]&0x7fffff)/4;
		DisplayError("JFG alInit\nmaxVVoices=%08x\nmaxPVoices=%08x\nmaxUpdates=%08x\nmaxFXbusses=%08x\nfxType=%02x\nparams=%08x",
			synconfig[0],synconfig[1],synconfig[2],synconfig[3],((BYTE*)synconfig)[(7*4+0)^3],synconfig[8]);

		if (((BYTE*)synconfig)[(7*4+0)^3]==6) { // custom fx
			FILE * params = fopen("h:\\fxcustom.txt","w");
			DWORD c,t,a=synconfig[8];
			r4300i_LW_VAddr(a,&c);
			r4300i_LW_VAddr(a+4,&t);
			fprintf(params,"{%ld, ",c); // count
			fprintf(params,"0x%lx,\n",t);

			a+=8;
			c*=8;
			c--;
			for (; c >= 0; c--,a+=4) {
				r4300i_LW_VAddr(a,&t);
				fprintf(params,"0x%x, ",t);
				if ((c%8)==0) fprintf(params,"\n");
				if (!c) break;
			}
			fprintf(params,"}");
			fclose(params);
		}
	}*/

	// JFG load of seq from ROM
	/*if (PROGRAM_COUNTER==0x8004D640 && GPR[4].UW[0]>=0x1d6de88 && GPR[4].UW[0] < 0x1d6de88+0x842f2) {
		//DisplayError("ra=%08x\na0=%08x\na1=%08x\na2=%08x",GPR[31].UW[0],GPR[4].UW[0],GPR[5].UW[0],GPR[6].UW[0]);
		ShowCPUStatus();
	}*/
	
	// All that's needed for JFG
	
	/*if (PROGRAM_COUNTER==0x80000510) {
		//DisplayError("a0=%08x\nra=%08x",GPR[4].UW[0],GPR[31].UW[0]);
		//GPR[4].UW[0]=0x55;
		//if (GPR[4].UW[0]==0x20) {
			GPR[4].UW[0]=0x55;
			//((DWORD*)N64MEM)[0x442A8/4]=0x03e00000;

			((DWORD*)N64MEM)[0x44C00/4]=0; // kill Thread2
			((DWORD*)N64MEM)[0x44BE8/4]=0; // kill Thread2 loop fcn 1 (no effect)
			((DWORD*)N64MEM)[0x44BF0/4]=0; // kill Thread2 loop fcn 2 (death)
			((DWORD*)N64MEM)[0x44BF8/4]=0; // kill Thread2 loop fcn 3 (no effect)

			((DWORD*)N64MEM)[0x4FAEC/4]=0; // kill Thread4 (lockup)

			((DWORD*)N64MEM)[0x66C94/4]=0; // kill Thread5
			((DWORD*)N64MEM)[0x66C7C/4]=0; // kill Thread5

			//((DWORD*)N64MEM)[0x9A260/4]=0; // kill Thread6 (needed for audio)
			//((DWORD*)N64MEM)[0x9A294/4]=0;
			//((DWORD*)N64MEM)[0x9A3CC/4]=0;
			//((DWORD*)N64MEM)[0x9A400/4]=0;

			//((DWORD*)N64MEM)[0x98CF0/4]=0; // kill Thread7
			//((DWORD*)N64MEM)[0x98D00/4]=0;

		//}
		sprintf(SaveAsFileName,"JFG");
		Machine_SaveState();
	}*/
	//if (PROGRAM_COUNTER==0x80000510) GPR[4].UW[0]=0x28;


	// Chopper Attack

	// disable chime on button press (alSndpPlay)
	/*if (PROGRAM_COUNTER==0x80009098) ((DWORD*)N64MEM)[0x9098/4]=0;

	if (PROGRAM_COUNTER==0x80007EC8) {
		DisplayError("PlaySong(%08x);\nra=%08x",GPR[4].UW[0],GPR[31].UW[0]);
		// kill Thread5 (no good)
		// kill Thread4 (voila)
		((DWORD*)N64MEM)[0xC1A4/4]=0;
		((DWORD*)N64MEM)[0xC18C/4]=0;

		sprintf(SaveAsFileName,"CA_OMFG");
		Machine_SaveState();
	}*/

	// Fighter's Destiny
	//if (PROGRAM_COUNTER==0x8025907C) DisplayError("a0=%08x\nra=%08x",GPR[4].UW[0],GPR[31].UW[0]);
	// 0x802eb3e0+0x84 = song to play
	// 0x802eb3e0+0x88 = current song
	
	// 0x802ed6b8+0x10c
	//if (PROGRAM_COUNTER==0x80256980) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x80256954) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x8025699C) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x8027030C) ShowCPUStatus();

	//if (PROGRAM_COUNTER==0x8025699C) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x8025BCA4) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x80259560) ShowCPUStatus();

	// Extreme G XG2

	//if (PROGRAM_COUNTER==0x8007BD08) DisplayError("thread %08x",GPR[6].UW[0]);
	//if (PROGRAM_COUNTER==0x8007EA68) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x8005C88C && GPR[4].UW[0]==0x15) {
	//if (PROGRAM_COUNTER==0x800876D8) DisplayError("osViaSwapBufer ra=%08x",GPR[31].UW[0]);

	/*if (PROGRAM_COUNTER==0x8005C88C) {
		DisplayError("song=%08x",GPR[4].UW[0]);
		GPR[4].UW[0]=0x11;
		//((DWORD*)N64MEM)[0x6082C/4]=0; // kill Thread7
		((DWORD*)N64MEM)[0x555D0/4]=0; // kill Thread6 (no effect?)
		((DWORD*)N64MEM)[0x4E4A0/4]=0; // kill Thread4

		((DWORD*)N64MEM)[0x56164/4]=0; // disable gfx message (huzzah!)
		//((DWORD*)N64MEM)[0x870F8/4]=0; // kill Thread2
		//((DWORD*)N64MEM)[0x870E8/4]=0; // kill Thread2
		//((DWORD*)N64MEM)[0x70B14/4]=0; // kill Thread8
		//((DWORD*)N64MEM)[0x8E294/4]=0; // Thread3 (needed for sound)

		sprintf(SaveAsFileName,"XG2_USF");
		Machine_SaveState();
	}*/

	// Paper Mario
	//if (PROGRAM_COUNTER==0x80065E60) DisplayError("thread %08x",GPR[6].UW[0]);
	//if (PROGRAM_COUNTER==0x8004AB00) DisplayError("SetupAudio ra=%08x",GPR[31].UW[0]);
	//if (PROGRAM_COUNTER==0x80056D5C) DisplayError("setmode? a1=%08x",GPR[5].UW[0]);
	//if (PROGRAM_COUNTER==0x80056A60) DisplayError("hi",GPR[4].UW[0]);
	/*if (PROGRAM_COUNTER==0x80061720
		&& GPR[31].UW[0] != 0x8005f4dc
		) DisplayError("osInvalDCache(%08x,%08x)\nra=%08x",GPR[4].UW[0],GPR[5].UW[0],GPR[31].UW[0]);
	*/

	//if (PROGRAM_COUNTER==0x80054E00) DisplayError("LoadFileFromROM\nROM=%08x\nRAM=%08x\nsize=%08x\nra=%08x",GPR[4].UW[0],GPR[5].UW[0],GPR[6].UW[0],GPR[31].UW[0]);

	//if (PROGRAM_COUNTER==0x8004B074) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x80058B78) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x80060E20 && GPR[31].UW[0] != 0x8004B1FC &&
	//	GPR[31].UW[0] != 0x8005F520 && GPR[31].UW[0] != 0x8006F234 && 
	//	GPR[31].UW[0] != 0x8006F5F4 && GPR[31].UW[0] != 0x8006F864) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x8004B9E4 && GPR[6].UW[0]==2) GPR[6].UW[0]=0x11; //ShowCPUStatus();
	
	// SONG SELECT:
	//if (PROGRAM_COUNTER==0x800558D4) /*GPR[4].UW[0]=0x6a;*/ ShowCPUStatus();

	// where we ultimately call down to SongSelect from

	//if (PROGRAM_COUNTER==0x8014A6AC) {
	/*if (PROGRAM_COUNTER==0x800558D4 && GPR[4].UW[0]==0x6a) {
		ShowCPUStatus();
		//GPR[4].UW[0]=0x95;
		
		// kill Thread9 (stops gfx, but game still running)
		((DWORD*)N64MEM)[0x2CB30/4]=0x8FBF0028; // load ra
		((DWORD*)N64MEM)[0x2CB34/4]=0x03E00008; // jr ra
		((DWORD*)N64MEM)[0x2CB38/4]=0x27BD0030; // fix stack pointer

		// kill Thread8 (no effect?)
		((DWORD*)N64MEM)[0x60300/4]=0x8FBF0030;
		((DWORD*)N64MEM)[0x60304/4]=0x03E00008;
		((DWORD*)N64MEM)[0x60308/4]=0x27BD0038;

		// Kill Thread7
		((DWORD*)N64MEM)[0x2C870/4]=0;

		// kill Thread10 (kills everything)
		//((DWORD*)N64MEM)[0x2D144/4]=0;
		//((DWORD*)N64MEM)[0x2D150/4]=0;
		//((DWORD*)N64MEM)[0x2D154/4]=0x8FBF0050;
		//((DWORD*)N64MEM)[0x2D158/4]=0x03E00008;
		//((DWORD*)N64MEM)[0x2D16C/4]=0x27BD0058;
		((DWORD*)N64MEM)[0x2D068/4]=0;
		((DWORD*)N64MEM)[0x2D13C/4]=0;
		((DWORD*)N64MEM)[0x2D11C/4]=0;
		((DWORD*)N64MEM)[0x2D098/4]=0;
		//disablecompare=1;

		// kill Thread3
		((DWORD*)N64MEM)[0x5F0B4/4]=0x8FBF002c;
		((DWORD*)N64MEM)[0x5F0B4/4]=0x03E00008;
		((DWORD*)N64MEM)[0x5F0B4/4]=0x27BD0030;

		// kill Thread2
		//((DWORD*)N64MEM)[0x66D80/4]=0;
		//((DWORD*)N64MEM)[0x66D70/4]=0;

		sprintf(SaveAsFileName,"PM_HITME");
		Machine_SaveState();
	}*/

	//if (PROGRAM_COUNTER==0x80053CF8) PROGRAM_COUNTER=0x80053E58;
	//if (PROGRAM_COUNTER==0x800558D4) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x8014A7E0) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x8014A918) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x8014AC3C) {
		//GPR[5].UW[0]=0x92;
		//GPR[6].UW[0]=0x01;
	//}
	//if (PROGRAM_COUNTER==0x8004D8B8) ShowCPUStatus();

	//if (PROGRAM_COUNTER==0x80053D3c) ShowCPUStatus();


	// NGE
	//if (PROGRAM_COUNTER==0x800B9EE4) DisplayError("alLoadNew ra=%08x",GPR[31].UW[0]);
	/*if (PROGRAM_COUNTER==0x8009B5C0) {
		//ShowCPUStatus();
		GPR[4].UW[0]=4;  // Zankokuna Tenshi no Thesis
		((DWORD*)N64MEM)[0xB6494/4]=0;

		sprintf(SaveAsFileName,"NGE_HITME");
		Machine_SaveState();
	}*/

	// Mischief Makers
	//if (PROGRAM_COUNTER==0x80099110) DisplayError("thread=%08x",GPR[6].UW[0]);

	/*if (PROGRAM_COUNTER==0x8009AEC4) {
		switch (((DWORD*)N64MEM)[(GPR[4].UW[0]&0x1fffffff)/4]) {
		case 1:
			PROGRAM_COUNTER=GPR[31].UW[0];
			break;
		case 2:
			//PROGRAM_COUNTER=GPR[31].UW[0];
			break;
		default:
			DisplayError("type=%08x",((DWORD*)N64MEM)[(GPR[4].UW[0]&0x1fffffff)/4]);
			break;
		}

	}*/

	/*if (PROGRAM_COUNTER==0x8009B110) {
		switch (((DWORD*)N64MEM)[(GPR[4].UW[0]&0x1fffffff)/4]) {
		case 1:
			//PROGRAM_COUNTER=GPR[31].UW[0];
			DisplayError("dlist from: %08x\n%08x",GPR[31].UW[0],GPR[4].UW[0]);
			break;
		case 2:
			//PROGRAM_COUNTER=GPR[31].UW[0];
			DisplayError("alist from %08x\n%08x",GPR[31].UW[0],GPR[4].UW[0]);
			break;
		default:
			DisplayError("type=%08x",((DWORD*)N64MEM)[(GPR[4].UW[0]&0x1fffffff)/4]);
			break;
		}

	}*/

	// here down was the code that worked
	/*
	if (PROGRAM_COUNTER==0x80002AD0) {
		GPR[4].UW[0]=2;
		((DWORD*)N64MEM)[0xA8118/4]=0; // kill Thread7, no effect?
		//((DWORD*)N64MEM)[0xA76E4/4]=0; // kill Thread5 (breaks sound)
		//((DWORD*)N64MEM)[0x99830/4]=0; // kill Thread4 (kills everything)
		//((DWORD*)N64MEM)[0xF0C/4]=0; // kill Thread3 (kills everything)
		((DWORD*)N64MEM)[0x9A2EC/4]=0; // kill Thread2, no effect?

		// things in Thread4
		((DWORD*)N64MEM)[0x996EC/4]=0; // no effect?
		//((DWORD*)N64MEM)[0x99718/4]=0; // needed
		//((DWORD*)N64MEM)[0x99740/4]=0; // interrupt handling?
		((DWORD*)N64MEM)[0x99818/4]=0; // not important?

		// things in Thread3
		((DWORD*)N64MEM)[0xd04/4]=0;
		((DWORD*)N64MEM)[0xd20/4]=0;
		
		((DWORD*)N64MEM)[0xe30/4]=0;
		((DWORD*)N64MEM)[0xe3c/4]=0; // vital interrupt thing
		//((DWORD*)N64MEM)[0xE44/4]=0; // needed for sound
		//((DWORD*)N64MEM)[0xE4C/4]=0; // DMAs fail
		
		((DWORD*)N64MEM)[0xe64/4]=0; // video not updated but still rendered
		
		((DWORD*)N64MEM)[0xe78/4]=0; // this guy causes us to wait for dlists
		//((DWORD*)N64MEM)[0xE80/4]=0; // no alists
		
		((DWORD*)N64MEM)[0xe88/4]=0; // no effect?
		((DWORD*)N64MEM)[0xe90/4]=0; // no effect?
		((DWORD*)N64MEM)[0xeb0/4]=0; // no effect?
		((DWORD*)N64MEM)[0xeb8/4]=0; // no effect?

		((DWORD*)N64MEM)[0xec8/4]=0;
		((DWORD*)N64MEM)[0xed0/4]=0;

		((DWORD*)N64MEM)[0xee0/4]=0;

		((DWORD*)N64MEM)[0xf04/4]=0;

		//ShowCPUStatus();

		sprintf(SaveAsFileName,"MM_HITME");
		Machine_SaveState();
	}
	*/

	// animal crossing

	//if (PROGRAM_COUNTER==0x8002FA50 && GPR[6].UW[0]==0x800d289c) DisplayError("thread=%08x ra=%08x",GPR[6].UW[0],GPR[31].UW[0]);
	//if (PROGRAM_COUNTER==0x800D4710) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x80035EF0) ShowCPUStatus();

	/*if (PROGRAM_COUNTER==0x800358E0 && shownewfunctioncalls) {
		if (((DWORD*)ROM)[GPR[6].UW[0]/4]!=0x59617A30) fprintf(loggy,"size=%08x RAM=%08x ROM=%08x\n",GPR[0xe].UW[0],GPR[7].UW[0],GPR[6].UW[0]);
		else fprintf(loggy,"Yaz0 size=%08x RAM=%08x ROM=%08x\n",GPR[0xe].UW[0],GPR[7].UW[0],GPR[6].UW[0]);

	}*/

	//if (PROGRAM_COUNTER==0x80026500 && GPR[31].UW[0] != 0x80026fd4) ShowCPUStatus();

	//if (PROGRAM_COUNTER==0x800271E8 && GPR[31].UW[0] != 0x800269a0) DisplayError("yaz0 a0=%08x a1=%08x a2=%08x ra=%08x",GPR[4].UW[0],GPR[5].UW[0],GPR[6].UW[0],GPR[31].UW[0]);

	//if (PROGRAM_COUNTER==0x800E6A34) ShowCPUStatus();

	// the power of DUMB LUCK triumphs!
	// song select: 8005DF70 a0=song a1=0x168?

	//if (PROGRAM_COUNTER==0x8005DF70) GPR[4].UW[0]=0x58;
	//if (PROGRAM_COUNTER==0x8005E10C) GPR[4].UW[0]=0x58;
	//if (PROGRAM_COUNTER==0x8005DBA4) {
		//if (GPR[6].UW[0]==7) PROGRAM_COUNTER==GPR[31].UW[0];
		//ShowCPUStatus();
	//}
	
	/*if (PROGRAM_COUNTER==0x8005F038) {
		PROGRAM_COUNTER=0x8005F064;
		//ShowCPUStatus();
		//((DWORD*)N64MEM)[0xD35E0/4]=0;
		//((DWORD*)N64MEM)[0xD35E8/4]=0;

		// kill Thread8 (needed)
		//for (cp=0xD8754; cp <= 0xD8788; cp+=4)
		//	((DWORD*)N64MEM)[cp/4]=0;		

		// kill Thread11 (...)
		//((DWORD*)N64MEM)[0xD46D0/4]=0;

		// kill Thread10
		//((DWORD*)N64MEM)[0xD74F0/4]=0;

		// kill Thread9
		((DWORD*)N64MEM)[0xD289C/4]=0;

		// kill Thread7 (freezes on a gfx frame)
		//((DWORD*)N64MEM)[0xD4C9C/4]=0;

		// kill Thread6
		((DWORD*)N64MEM)[0x297E0/4]=0;

		// kill Thread2 (freezes on a gfx frame)
		//((DWORD*)N64MEM)[0x3AAC8/4]=0;
	}

	if (PROGRAM_COUNTER==0x8005DF70) {
		((DWORD*)N64MEM)[0x5F068/4]=0; // disable subsequent calls
		//ShowCPUStatus();
		//GPR[4].UW[0]=0x2c; //0x2b;
		//GPR[5].UW[0]=0x200;

		//((DWORD*)N64MEM)[0xD26D8/4]=0; // disable audio
		//((DWORD*)N64MEM)[0xD41DC/4]=0;
		((DWORD*)N64MEM)[0xD875C/4]=0; // works briefly...
		((DWORD*)N64MEM)[0xD876C/4]=0;
		((DWORD*)N64MEM)[0xD877C/4]=0;

		((DWORD*)N64MEM)[0x29AF4/4]=0; // these weird ops kill progress
		((DWORD*)N64MEM)[0x29AF8/4]=0;

		//((DWORD*)N64MEM)[0x5D084/4]=0; // freeze
		//((DWORD*)N64MEM)[0x5D13C/4]=0;

		sprintf(SaveAsFileName,"AC_BLAH");
		Machine_SaveState();
	}*/

	// other fcn, looks a lot like SongSelect	
	//if (PROGRAM_COUNTER==0x8005DD2C) GPR[4].UW[0]=0;
	// ditto
	//if (PROGRAM_COUNTER==0x8005DC9C) GPR[4].UW[0]=0;
	// this is responsible for writing the song#
	//if (PROGRAM_COUNTER==0x8005DBA4) ShowCPUStatus();
	
	//if (PROGRAM_COUNTER==0x8005ACC0) ShowCPUStatus(); //PROGRAM_COUNTER=GPR[31].UW[0];
	
	//if (PROGRAM_COUNTER==0x800D35E0) ShowCPUStatus();
	
	//if (PROGRAM_COUNTER==0x800924CC) PROGRAM_COUNTER=GPR[31].UW[0]; // removes pause

	//if (PROGRAM_COUNTER==0x8005DC00) ShowCPUStatus();
		
	// conker's bad fur day

	//if (PROGRAM_COUNTER==0x80022A60) DisplayError("osStartThread ra=%08x",GPR[31].UW[0]);

	// Extreme G

	//if (PROGRAM_COUNTER==0x8007E550) DisplayError("thread=%08x",GPR[6].UW[0]);
	//if (PROGRAM_COUNTER==0x800808BC) DisplayError("alCSeqNew ra=%08x",GPR[31].UW[0]);

	/*if (PROGRAM_COUNTER==0x800591C0) {
		GPR[4].UW[0]=1;
		((DWORD*)N64MEM)[0x591cc/4]=0x3C018009; // read song from a0, not v1
		((DWORD*)N64MEM)[0x591d0/4]=0xAC245D2C; // "

		((DWORD*)N64MEM)[0x54570/4]=0; // eliminate gfx mesg

		sprintf(SaveAsFileName,"XG_USF");
		Machine_SaveState();
	}*/

	// Mario Party
	//if (PROGRAM_COUNTER==0x8008B650) ShowCPUStatus();
	/*if (PROGRAM_COUNTER==0x8000BEBC) {
		DisplayError("blah, a0=%08x",GPR[4].UW[0]);
		GPR[4].UW[0]=0x5;
		((DWORD*)N64MEM)[0x1A970/4]=0;
		((DWORD*)N64MEM)[0x1A9C4/4]=0;

		sprintf(SaveAsFileName,"MP1_2_USF");
		Machine_SaveState();
	}*/


	//if (PROGRAM_COUNTER==0x8000B9E4) GPR[2].UW[0]=0x4f;

	//if (PROGRAM_COUNTER==0x800d18c4) ShowCPUStatus();

	// Conker's Bad Fur Day (the trick is to make a save state after the intro)

	/*if (TLBTrans(PROGRAM_COUNTER)==0xD96C) {
		//DisplayError("song select %04x ra=%08x (%08x)",GPR[4].UW[0],GPR[31].UW[0],TLBTrans(GPR[31].UW[0]));
		//ShowCPUStatus();
		//DisplayError("ra=%08x",TLBTrans(GPR[31].UW[0]));
		GPR[4].UW[0]=0x42;
		GPR[6].UW[0]=5;
		((DWORD*)N64MEM)[0x507C/4]=0;
		((DWORD*)N64MEM)[0x5030/4]=0;
		//((DWORD*)N64MEM)[0x3740/4]=0; //nope
		((DWORD*)N64MEM)[0x4fc8/4]=0;
		
		//((DWORD*)N64MEM)[0x3790/4]=0; // kill Thread5
		//((DWORD*)N64MEM)[0x37A0/4]=0;

		((DWORD*)N64MEM)[0x13DC/4]=0; // kill Thread3 (no effect?)

		//((DWORD*)N64MEM)[0x3058/4]=0; // kill Thread2 (never gets to music)
		//((DWORD*)N64MEM)[0x3060/4]=0;

		//((DWORD*)N64MEM)[0x3034/4]=0;
		//((DWORD*)N64MEM)[0x3044/4]=0;

		((DWORD*)N64MEM)[0x308F84/4]=0;
		((DWORD*)N64MEM)[0x30A5B0/4]=0;

		((DWORD*)N64MEM)[0x308BB8/4]=0x03E00008;
		((DWORD*)N64MEM)[0x308BBC/4]=0;

		//used=1;

		//sprintf(SaveAsFileName,"MIGHTYPOO_USF");
		//Machine_SaveState();
	}*/

	/*if (TLBTrans(PROGRAM_COUNTER)==0x24410 && used) {
		used++;
		if (used==50) {((DWORD*)N64MEM)[0x4FC8/4]=0; used=0;}
	}*/

	//if (TLBTrans(PROGRAM_COUNTER)==0x37F0) DisplayError("thread=%08x",TLBTrans(GPR[6].UW[0]));
	//if (TLBTrans(PROGRAM_COUNTER)==0x2E2520) ShowCPUStatus();
	//if (((WORD*)N64MEM)[0x317ca/2]==0x0022) ShowCPUStatus();
	
	// Shadows of the Empire

	//if (TLBTrans(PROGRAM_COUNTER)==0xBF038) ShowCPUStatus();

	// Bomberman 64
	//if (PROGRAM_COUNTER==0x8000843C) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x80002730) DisplayError("thread=%08x",GPR[6].UW[0]);
	/*if (PROGRAM_COUNTER==0x8026C93C) {
		ShowCPUStatus();
		//GPR[4].UW[0]=0x2;
		((DWORD*)N64MEM)[0x2380C8/4]=0; // kill Thread6 (that'll do it)
		((DWORD*)N64MEM)[0x226D34/4]=0; // kill Thread7 (good, stops gfx, but controller still read)

		sprintf(SaveAsFileName,"BOMBMAN_USF");
		Machine_SaveState();
	}*/

	//if (PROGRAM_COUNTER==0x80005E40) ShowCPUStatus();

	// Yoshi's Story

	//if (PROGRAM_COUNTER==0x80085B10) DisplayError("thread=%08x",GPR[6].UW[0]);
	//if (PROGRAM_COUNTER==0x8008BE00) DisplayError("osAiSetFreq ra=%08x",GPR[31].UW[0]);

	// Turok

	//if (PROGRAM_COUNTER==0x8009DF50) DisplayError("thread=%08x",GPR[6].UW[0]);
	//if (PROGRAM_COUNTER==0x800A456C) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x8008B57C) ShowCPUStatus();
	//if (GPR[5].UW[0]==0x802e3e60) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x8002D5E0 && GPR[31].UW[0]==0x8008B568) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x80000A28) ShowCPUStatus();

	// Rogue Squadron

	//if (PROGRAM_COUNTER==0x80032420 && GPR[31].UW[0]!=0x800035e8 && GPR[31].UW[0]!=0x800037c4) ShowCPUStatus();
	/*if (Opcode.op==3 && !doneyet[(JumpToLocation&0x3fffff)/4]) {
		if (GPR[4].UW[0] != 0) {
			doneyet[(JumpToLocation&0x3fffff)/4]=1;
			if (GPR[4].UW[0] <= 0x60) {
				if (!fcnlog) fcnlog=fopen("c:\\fcnlog.txt","w");
				fprintf(fcnlog,"%08x a0=%08x\n",JumpToLocation,GPR[4].UW[0]);
			}
		}
	}*/
	/*if (PROGRAM_COUNTER==0x80066620) {
		if (GPR[6].UW[0]==0x8003C2F8) ShowCPUStatus();
		//GPR[4].UW[0]=0x91;
	}*/

	//((BYTE*)N64MEM)[0xA0F50^3]=1;

	//for (c=1; c < 32; c++) if (GPR[c].UW[0]==0x8006657C) {ShowCPUStatus(); break;}
	//if (PROGRAM_COUNTER==0x80066918) ShowCPUStatus();
	/*if (PROGRAM_COUNTER==0x800A9668) {
		((WORD*)N64MEM)[(0xaf810^2)/2]=0x2;

		sprintf(SaveAsFileName,"RS_MAINTHEME");
		Machine_SaveState();

		//((DWORD*)N64MEM)[0x19E9C/4]=0;
		//((DWORD*)N64MEM)[0x19F1C/4]=0;
		
	}*/

	// SF Rush
	//if (PROGRAM_COUNTER==0x80006F30) ShowCPUStatus(); // threads
	/*if (PROGRAM_COUNTER==0x800979A0) {
		GPR[4].UW[0]=1;

		// Kill Thread12 (good!)
		((DWORD*)N64MEM)[0x251C/4]=0;

		// Kill Thread13
		//((DWORD*)N64MEM)[0x8A9E4/4]=0;

		// Kill Thread11
		//((DWORD*)N64MEM)[0x99AD8/4]=0;

		// Kill Thread10
		//((DWORD*)N64MEM)[0x10C30/4]=0;

		sprintf(SaveAsFileName,"SFR_SS");
		Machine_SaveState();
	}*/

	// Pocket Monsters Stadium (J)
	//if (PROGRAM_COUNTER==0x80051B20) DisplayError("thread=%08x",GPR[6].UW[0]);
	//if (PROGRAM_COUNTER==0x8004609C) {
	//	if (!dmalog) dmalog=fopen("c:\\dmalog.txt","w");
	//	fprintf(dmalog,"a0=%08x a1=%08x a2=%08x ra=%08x\n",GPR[4].UW[0],GPR[5].UW[0],GPR[6].UW[0],GPR[31].UW[0]);
	//}
	/*if (PROGRAM_COUNTER==0x80003EB8) {
		//ShowCPUStatus();
		GPR[5].UW[0]=0x3F53F0;
		GPR[6].UW[0]=0x3FC1A0;
	}*/
	//if (PROGRAM_COUNTER==0x80116880) {
		//sprintf(SaveAsFileName,"PS1");
		//Machine_SaveState();
	//}

	/*if (PROGRAM_COUNTER==0x8004609C) {
		//ShowCPUStatus();


		
		if (GPR[4].UW[0]==0xC5B680) {
			int c = 0x14;
			GPR[4].UW[0]=((DWORD*)N64MEM)[0xB984C/4+c];
			GPR[6].UW[0]=((DWORD*)N64MEM)[0xB984C/4+c+1]-((DWORD*)N64MEM)[0xB984C/4+c];
			if (c==0x14) GPR[6].UW[0]=0x1f0; // a "next" seq isn't in the table, but this
				// is where the last sequence (a bit of fanfare) ends

			// kill Thread8
			((DWORD*)N64MEM)[0x181C/4]=0;

			sprintf(SaveAsFileName,"PS1_MOOCOW");
			Machine_SaveState();
		}
	}*/

	
	/*if (PROGRAM_COUNTER==0x8003F5A0) {
		// when the selection screen comes up this word (in stack) controls song
		((DWORD*)N64MEM)[(0x82a30+0x24)/4]=0x0; //ShowCPUStatus();

		// kill Thread8
		//((DWORD*)N64MEM)[0x181C/4]=0;
	}*/

	// Pocket Monsters Stadium 2 (J)
	//if (PROGRAM_COUNTER==0x800558E0) DisplayError("Thread=%08x",GPR[6].UW[0]);
	//if (PROGRAM_COUNTER==0x80047C70) ShowCPUStatus();
	//if (PROGRAM_COUNTER==0x80048744) GPR[9].UW[0]=1;
	/*if (PROGRAM_COUNTER==0x80047C70) {
		if (GPR[4].UW[0]==0x16d0ac0) {
			int c = 0x48;
			//GPR[4].UW[0]=((DWORD*)N64MEM)[0xE281c/4+c];
			//GPR[6].UW[0]=((DWORD*)N64MEM)[0xE281c/4+c+1]-((DWORD*)N64MEM)[0xE281c/4+c];
			//if (c==0x48) GPR[6].UW[0]=0x7d4;
			//GPR[4].UW[0]=0x16f1c14;
			//GPR[6].UW[0]=0x360;

			// kill Thread8
			((DWORD*)N64MEM)[0x1928/4]=0;

			sprintf(SaveAsFileName,"PS2_MOOCOW");
			Machine_SaveState();
		}
	}*/

	// Pocket Monsters Stadium GS (J)
	//if (PROGRAM_COUNTER==0x80078A60) DisplayError("thread=%08x",GPR[6].UW[0]);
	//if (PROGRAM_COUNTER==0x800220B0) GPR[16].UW[0]=0x01; //0x24; //ShowCPUStatus();
	// song <= 0x27, >= 0x88 causes instant end-of-song condition, demo starts
	// song <= 0, >= 0x62 don't work
	/*if (PROGRAM_COUNTER==0x80022090 && GPR[16].UW[0]==0x25) {
		GPR[16].UW[0]=0x61;
		//ShowCPUStatus();

		// kill Thread9
		((DWORD*)N64MEM)[0x1C70/4]=0;

		// kill Thread5
		((DWORD*)N64MEM)[0x7AD98/4]=0;

		sprintf(SaveAsFileName,"PSGS_MOOCOW");
		Machine_SaveState();
	}*/

	// S

	if (!r4300i_LW_VAddr(PROGRAM_COUNTER, &Opcode.Hex)) { 
		DoTLBMiss(NextInstruction == JUMP,PROGRAM_COUNTER);
		NextInstruction = NORMAL;
		return;
	}

	COUNT_REGISTER += CountPerOp;
	if (CPU_Type != CPU_SyncCores) { Timers.Timer -= CountPerOp; }

	RANDOM_REGISTER -= 1;
	if ((int)RANDOM_REGISTER < (int)WIRED_REGISTER) {
		RANDOM_REGISTER = 31;
	}

	((void (_fastcall *)()) R4300i_Opcode[ Opcode.op ])();
	if (GPR[0].DW != 0) {
#if (!defined(EXTERNAL_RELEASE))
		DisplayError("GPR[0].DW has been written to");
#endif
		GPR[0].DW = 0;
	}

	switch (NextInstruction) {
	case NORMAL: 
		PROGRAM_COUNTER += 4; 
		break;
	case DELAY_SLOT:
		NextInstruction = JUMP;
		PROGRAM_COUNTER += 4; 
		break;
	case JUMP:
		PROGRAM_COUNTER  = JumpToLocation;
		/*
		if (GPR[5].UW[0]==0x800B91A0
			&& PROGRAM_COUNTER!=0x80056250
			&& PROGRAM_COUNTER!=0x80056290
			&& PROGRAM_COUNTER!=0x800566a8
			&& PROGRAM_COUNTER!=0x800533d0
			&& PROGRAM_COUNTER!=0x80056cec
			&& PROGRAM_COUNTER!=0x8004b014
			&& PROGRAM_COUNTER!=0x8004b024
			&& PROGRAM_COUNTER!=0
			) DisplayError("fcn called with AudioObject\npc=%08x",PROGRAM_COUNTER);
		*/
		NextInstruction = NORMAL;
		if (CPU_Type != CPU_SyncCores) {
			if (Profiling) {
				if (IndvidualBlock) {
					char Label[100];
					sprintf(Label,"PC: %X",PROGRAM_COUNTER);
					StartTimer(Label);
				} else {
					StartTimer("r4300i Running");
				}
			}
			if (CPU_Type != CPU_SyncCores) {
				if ((int)Timers.Timer < 0) {  TimerDone(); }
				if (CPU_Action.DoSomething) { DoSomething(); }
			}
		}
	}		
}
	
void StartInterpreterCPU (void ) { 
	//DWORD Value, Value2, Addr = 0x80031000;
	int c;

	//loggy=fopen("c:\\pilog.txt","w");

	//for (c=0; c < 0x100000; c++) fcnused[c]=0;

#ifdef AL_USE
	LoadLibrarySignatures();
#endif

	NextInstruction = NORMAL;

	if (GfxRomOpen != NULL) { GfxRomOpen(); }
	if (ContRomOpen != NULL) { ContRomOpen(); }
	//Add_R4300iBPoint(0x802000C8,FALSE);
	__try {
		for (;;) {
#if (!defined(EXTERNAL_RELEASE))
			if (NoOfBpoints != 0) {
				if (CheckForR4300iBPoint(PROGRAM_COUNTER)) {
					UpdateCurrentR4300iRegisterPanel();
					Refresh_Memory();
					if (InR4300iCommandsWindow) {
						Enter_R4300i_Commands_Window();
						SetR4300iCommandViewto( PROGRAM_COUNTER );
						if (CPU_Action.Stepping) {
							DisplayError ( "Encounted a R4300i Breakpoint" );
						} else {
							DisplayError ( "Encounted a R4300i Breakpoint\n\nNow Stepping" );
							SetR4300iCommandToStepping();
						}
					} else {
						DisplayError ( "Encounted a R4300i Breakpoint\n\nEntering Command Window" );
						Enter_R4300i_Commands_Window();
					}					
				}
			}

			//r4300i_LW_VAddr(Addr,&Value);
			//if (Value2 != Value) {
			//	DisplayError("%X changed",Addr);
			//}
			//Value2 = Value;
			if (CPU_Action.Stepping) {
				do {
					SetR4300iCommandViewto (PROGRAM_COUNTER);
					UpdateCurrentR4300iRegisterPanel();
					Refresh_Memory();
					WaitForSingleObject( CPU_Action.hStepping, INFINITE );
					if (CPU_Action.Stepping) { ExecuteInterpreterOpCode(); }
				} while (CPU_Action.Stepping);
			}
#endif
			ExecuteInterpreterOpCode();
		}
	} __except( r4300i_CPU_MemoryFilter( GetExceptionCode(), GetExceptionInformation()) ) {
		DisplayError("Unknown memory action\n\nEmulation stop");
		ExitThread(0);
	}
}

void TestInterpreterJump (DWORD PC, DWORD TargetPC, int Reg1, int Reg2) {
	if (PC != TargetPC) { return; }
	if (DelaySlotEffectsCompare(PC,Reg1,Reg2)) { return; }
	if (CPU_Type != CPU_Interpreter) { return; }
	InPermLoop();
}

